% tzconv  convert between timezones
%
% Syntax:
%   d = tzconv(dates, from_zone, to_zone);
%
% Input Arguments:
%   dates     - a vector of matlab dates to be converted
%   from_zone - interpret DATES as coming from FROM_ZONE
%   to_zone   - convert DATEs to TO_ZONE
%
% Output Arguments:
%   d      - vector of dates in the result
%
% Copyright (C) 2019 Sebastian Freundt <freundt@ga-group.nl>
%
% This file is part of dateutils
